-- textmap.e
-- Pete Eberlein <xseal@harborside.com>
-- polygon texture mapping include file

---------------------------<setup_text_map>-----------------------------------------
-- 
-- Syntax:      setup_text_map(screen, width, height, textures, clear_color)
-- 
-- Description: Sets up the machine codes for text_map.
--              screen is an atom describing the address the screen.
--              width is an integer describing the width of the screen.
--              height is an integer describing the height of the screen.
--              textures is an atom describing the address of a 256 byte wide
--              bitmap in memory containing all available textures.
--              clear_color lets you define the color that is see-through.

------------------------------<text_map>-----------------------------------------
-- 
-- Syntax:      text_map(bitmap, vertices)
-- 
-- Description: Maps a bitmap onto a polygon defined by vertices.
--              bitmap is a length-4 sequence of length-2 sequences of
--              the form {x,y}.  
--              vertices is a length-4 sequence of length-2 sequences of 
--              the form {x,y}.
-- 
-- Comments:    The polygon is clipped to the bounds of the screen, defined
--              by width and height in setup_text_map.  Values passed to bitmap
--              must be in fixed point integers: whole_number * #10000 + fractional.
--              The bitmaps vertices should be calculated carefully to avoid 
--              going out of the bounds of the texture bitmap.  Due to rounding
--              errors, the procedure sometimes goes outside the bitmap vertices,
--              so put a clear color buffer around the bitmap region.  Clear
--              color defined by setup_text_map is see-through.
--              
    

include machine.e

constant vect_mem = allocate(48*4),
         text_map_proc = allocate(371)

function get_slope(sequence v1, sequence v2)
    atom height
    atom d
    sequence dx
    height = v2[2] - v1[2]
    if height = 0 then 
        dx = repeat(0, length(v1))
    else
        dx = floor(#10000 * (v2 - v1) / height)
        d = #FFFF - and_bits(v1[2]+#FFFF, #FFFF)
        v1 = floor(v1 + dx * d / #10000)
    end if
    return v1 & dx
end function

global procedure text_map(sequence bitmap, sequence vertices)
-- maps a bitmap onto a four-sided polygon
--  bitmap is a length-4 sequence of bitmap corners
--  vertices is a length-4 sequence of length-2 sequences
    sequence vectdata, tm_data
    vectdata = repeat(0, 48)
    tm_data = repeat(0, 192)
    for i = 1 to 4 do
        vertices[i] = vertices[i] & bitmap[i]
    end for
    while (vertices[1][2] > vertices[4][2])
       or (vertices[1][2] > vertices[3][2])
       or (vertices[1][2] > vertices[2][2]) do
        vertices = append(vertices[2..4], vertices[1])
    end while
    if (vertices[2][1]-vertices[1][1])*(vertices[4][2]-vertices[1][2]) > 
       (vertices[4][1]-vertices[1][1])*(vertices[2][2]-vertices[1][2]) then
        vertices = {vertices[1], vertices[4], vertices[3], vertices[2]}
    end if
    vectdata[1..8] = get_slope(vertices[1], vertices[2])
    vectdata[21..28] = get_slope(vertices[1], vertices[4])
    vectdata[20] = vect_mem + 32
    if vertices[2][2] < vertices[4][2] then
        vectdata[19] = floor((vertices[2][2]+#FFFF) / #10000) - floor((vertices[1][2]+#FFFF) / #10000)
        vectdata[31..38] = get_slope(vertices[2], vertices[3])
        vectdata[30] = vect_mem
        if vertices[3][2] < vertices[4][2] then
            vectdata[29] = floor((vertices[3][2]+#FFFF) / #10000) - floor((vertices[2][2]+#FFFF) / #10000)
            vectdata[41..48] = get_slope(vertices[3], vertices[4])
            vectdata[39] = floor((vertices[4][2]+#FFFF) / #10000) - floor((vertices[3][2]+#FFFF) / #10000)
            vectdata[40] = vect_mem
        else
            vectdata[29] = floor((vertices[4][2]+#FFFF) / #10000) - floor((vertices[2][2]+#FFFF) / #10000)
            vectdata[41..48] = get_slope(vertices[4], vertices[3])
            vectdata[39] = floor((vertices[3][2]+#FFFF) / #10000) - floor((vertices[4][2]+#FFFF) / #10000)
            vectdata[40] = vect_mem + 32
        end if
    else
        vectdata[19] = floor((vertices[4][2]+#FFFF) / #10000) - floor((vertices[1][2]+#FFFF) / #10000)
        vectdata[31..38] = get_slope(vertices[4], vertices[3])
        vectdata[30] = vect_mem + 32
        if vertices[3][2] < vertices[2][2] then
            vectdata[29] = floor((vertices[3][2]+#FFFF) / #10000) - floor((vertices[4][2]+#FFFF) / #10000)
            vectdata[41..48] = get_slope(vertices[3], vertices[2])
            vectdata[39] = floor((vertices[2][2]+#FFFF) / #10000) - floor((vertices[3][2]+#FFFF) / #10000)
            vectdata[40] = vect_mem + 32
        else
            vectdata[29] = floor((vertices[2][2]+#FFFF) / #10000) - floor((vertices[4][2]+#FFFF) / #10000)
            vectdata[41..48] = get_slope(vertices[2], vertices[3])
            vectdata[39] = floor((vertices[3][2]+#FFFF) / #10000) - floor((vertices[2][2]+#FFFF) / #10000)
            vectdata[40] = vect_mem
        end if
    end if        
    
    for i = 1 to length(vectdata) do
        tm_data[i*4-3..i*4] = floor(vectdata[i] / {1,256,65536,16777216})
    end for
    poke(vect_mem, tm_data)
    call(text_map_proc)
end procedure

         

global procedure setup_text_map(atom dest, integer width, integer height, atom bitmaps, integer clear_color)
    sequence vm, x1, y1, bx1, by1, x2, bx2, by2, d_bx, d_by,
        b_width, b_dest, b_dest_end, b_bitmaps
    if not (text_map_proc and vect_mem) then
        puts(1, "Not enough memory to allocate text_map procedure\n")
        abort(1)
    end if
    vm = int_to_bytes(vect_mem+72)
    x1 = int_to_bytes(vect_mem)
    y1 = int_to_bytes(vect_mem+4)
    bx1 = int_to_bytes(vect_mem+8)
    by1 = int_to_bytes(vect_mem+12)
    x2 = int_to_bytes(vect_mem+32)
    bx2 = int_to_bytes(vect_mem+40)
    by2 = int_to_bytes(vect_mem+44)
    d_bx = int_to_bytes(vect_mem+64)
    d_by = int_to_bytes(vect_mem+68)
    b_width = int_to_bytes(width)
    b_dest = int_to_bytes(dest)
    b_dest_end = int_to_bytes(dest+width*height)
    b_bitmaps = int_to_bytes(bitmaps)

poke(text_map_proc,
   {#60,                    --pusha
    #FC,                    --cld
    #BE} & vm & {           --mov esi, vect_mem (3)
    #B9,#03,#00,#00,#00,    --mov ecx, 3
    #A1} & y1 & {           --mov eax, [y1] (13)
    #05,#FF,#FF,#00,#00,    --add eax, #FFFF
    #C1,#F8,#10,            --sar eax, 16
    #BA} & b_width & {      --mov edx, line_length (26)
    #F7,#EA,                --imul edx
    #05} & b_dest & {       --add eax, dest_start (33)
    #89,#C7,                --mov edi, eax
    #51,                    --top: push ecx
    #57,                    --push edi
    #AD,                    --lodsd
    #89,#C1,                --mov ecx, eax
    #AD,                    --lodsd
    #89,#C7,                --mov edi, eax
    #A5,                    --movsd
    #A5,                    --movsd
    #A5,                    --movsd
    #A5,                    --movsd
    #A5,                    --movsd
    #A5,                    --movsd
    #A5,                    --movsd
    #A5,                    --movsd
    #5F,                    --pop edi
    #50,                    --push eax
    #21,#C9,                --and ecx, ecx
    #75,#06,                --jnz middle
    #58,                    --into_top: pop eax
    #59,                    --pop ecx
    #E2,#E6,                --loop top
    #61,                    --popa
    #C3,                    --ret
    #81,#FF} & b_dest & {   --middle: cmp edi, dest_start (69)
    #72,#1C,                --jb  in_the_middle
    #81,#FF} & b_dest_end &{--cmp edi, dest_end (77)
    #73,#14,                --jae in_the_middle
    #8B,#15} & x2 & {       --mov edx, [x2] (85)
    #2B,#15} & x1 & {       --sub edx, [x1] (91)
    #74,#06,                --jz in_the_middle
    #51,                    --push ecx
    #89,#D1,                --mov ecx, edx
    #EB,#35,                --jmp middle2
    #59,                    --back_into_middle: pop ecx
    #BA} & x1 & {           --in_the_middle: mov edx, x1 (104)
    #8B,#42,#10,            --mov eax,[edx+16]
    #01,#02,                --add [edx], eax
    #8B,#42,#18,            --mov eax,[edx+24]
    #01,#42,#08,            --add [edx+8], eax
    #8B,#42,#1C,            --mov eax,[edx+28]
    #01,#42,#0C,            --add [edx+12], eax
    #8B,#42,#30,            --mov eax,[edx+48]
    #01,#42,#20,            --add [edx+32], eax
    #8B,#42,#38,            --mov eax,[edx+56]
    #01,#42,#28,            --add [edx+40], eax
    #8B,#42,#3C,            --mov eax,[edx+60]
    #01,#42,#2C,            --add [edx+44], eax
    #81,#C7} & b_width & {  --add edi, line_length (145)
    #E2,#AC,                --loop middle
    #EB,#A4,                --jmp into_top
    #EB,#CB,                --into_middle: jmp back_into_middle
    #A1} & bx2 & {          --middle2: mov eax, [bx2] (156)
    #2B,#05} & bx1 & {      --sub eax, [bx1] (162)
    #89,#C2,                --mov edx, eax
    #C1,#FA,#10,            --sar edx, 16
    #C1,#E0,#10,            --shl eax, 16
    #F7,#F9,                --idiv ecx
    #A3} & d_bx & {         --mov [d_bx], eax (177)
    #A1} & by2 & {          --mov eax, [by2] (182)
    #2B,#05} & by1 & {      --sub eax, [by1] (188)
    #89,#C2,                --mov edx, eax
    #C1,#FA,#10,            --sar edx, 16
    #C1,#E0,#10,            --shl eax, 16
    #F7,#F9,                --idiv ecx
    #A3} & d_by & {         --mov [d_by], eax (203)
    #A1} & x1 & {           --mov eax, [x1] (208)
    #05,#FF,#FF,#00,#00,    --add eax, #FFFF
    #89,#C1,                --mov ecx, eax
    #25,#FF,#FF,#00,#00,    --and eax, #FFFF
    #F7,#D8,                --neg eax
    #05,#FF,#FF,#00,#00,    --add eax, #FFFF
    #C1,#F9,#10,            --sar ecx, 16
    #83,#F9,#00,            --cmp ecx, 0
    #7D,#09,                --jge x1_ok
    #F7,#D9,                --neg ecx
    #C1,#E1,#10,            --shl ecx, 16
    #01,#C8,                --add eax, ecx
    #31,#C9,                --xor ecx, ecx
    #89,#C2,                --x1_ok: mov edx, eax
    #A1} & x2 & {           --mov eax, [x2] (251)
    #05,#FF,#FF,#00,#00,    --add eax, #FFFF
    #C1,#F8,#10,            --sar eax, 16
    #3D} & b_width & {      --cmp eax, line_length (264)
    #7E,#05,                --jle x2_ok
    #B8} & b_width & {      --mov eax, line_length (271)
    #39,#C1,                --x2_ok: cmp ecx, eax
    #7C,#02,                --jl inner
    #EB,#80,                --into_middle2: jmp into_middle
    #57,                    --inner: push edi
    #01,#CF,                --add edi, ecx
    #F7,#D9,                --neg ecx
    #01,#C1,                --add ecx, eax
    #A1} & d_by & {         --mov eax, [d_by] (289)
    #52,                    --push edx
    #F7,#EA,                --imul edx
    #0F,#AC,#D0,#10,        --shrd eax, edx, 16
    #03,#05} & by1 & {      --add eax, [by1] (302)
    #89,#C5,                --mov ebp, eax
    #5A,                    --pop edx
    #A1} & d_bx & {         --mov eax, [d_bx] (310)
    #F7,#EA,                --imul edx
    #0F,#AC,#D0,#10,        --shrd eax, edx, 16
    #03,#05} & bx1 & {      --add eax, [bx1] (322)
    #89,#C3,                --mov ebx, eax
    #89,#D8,                --innermost: mov eax, ebx
    #C1,#E8,#10,            --shr eax, 16
    #89,#EA,                --mov edx, ebp
    #C1,#EA,#08,            --shr edx, 8
    #88,#C2,                --mov dl, al
    #8A,#82} & b_bitmaps & {--mov al, [edx + bitmaps] (342)
    #3C,clear_color,        --cmp al, clear_color
    #74,#03,                --jz clearcolor
    #AA,                    --stosb
    #EB,#01,                --jmp noclear
    #47,                    --clearcolor: inc edi
    #03,#2D} & d_by & {     --noclear: add ebp, [d_by] (356)
    #03,#1D} & d_bx & {     --add ebx, [d_bx] (362)
    #E2,#D8,                --loop innermost
    #5F,                    --pop edi
    #EB,#A4})               --jmp into_middle2
end procedure

